use std::{
    cmp::Ordering,
    ffi::OsStr,
    fs,
    path::{Path, PathBuf},
};

use chrono::{DateTime, Utc};
use log::debug;
use strum_macros::{EnumDiscriminants, EnumIter};

pub trait PdfSheet {
    fn get_pdf(&self) -> &Pdf;
}

#[derive(Debug, Clone, PartialEq, Eq)]
pub struct Sheet {
    pub id: i64,
    pub last_opened: I64DateTime,
    pub kind: SheetKind,
    pub pdf: Pdf,
}

#[derive(Debug, Clone, PartialEq, Eq, EnumDiscriminants)]
#[strum_discriminants(derive(EnumIter))]
pub enum SheetKind {
    Sheet {
        name: String,
        composer_id: i64,
        first_page: i64,
        book_id: Option<i64>,
    },
    Orphan,
    Book {
        name: String,
        composer_id: i64,
        sheet_ids: Vec<i64>,
    },
}

impl PartialOrd for Sheet {
    fn partial_cmp(&self, other: &Self) -> Option<Ordering> {
        Some(self.cmp(other))
    }
}

impl Ord for Sheet {
    fn cmp(&self, other: &Self) -> Ordering {
        self.last_opened.cmp(&other.last_opened)
    }
}

impl Sheet {
    pub fn construct_xopp_file_path(&self) -> PathBuf {
        let mut xopp_path = self.pdf.path.with_extension("").into_os_string();
        xopp_path.push(".xopp");
        PathBuf::from(xopp_path)
    }

    pub fn construct_annotated_file_path(&self) -> PathBuf {
        let mut annotated_path = self.pdf.path.with_extension("").into_os_string();
        annotated_path.push("_annotated.pdf");
        PathBuf::from(annotated_path)
    }

    pub fn open_file_or_annotated_version_if_exists(&self) {
        let annotated_version = self.construct_annotated_file_path();
        if annotated_version.exists() {
            // TODO: open on first_page
            opener::open(annotated_version).unwrap();
        } else {
            // TODO: open on first_page
            opener::open(&self.pdf.path).unwrap();
        }
    }

    pub fn is_part_of_book(&self) -> bool {
        if let SheetKind::Sheet { book_id, .. } = &self.kind {
            return book_id.is_some();
        }
        false
    }
}

impl SheetKindDiscriminants {
    pub fn get_database_table_name(&self) -> &str {
        match self {
            SheetKindDiscriminants::Sheet => "sheets",
            SheetKindDiscriminants::Orphan => "orphans",
            SheetKindDiscriminants::Book => "books",
        }
    }
}

#[derive(sqlx::FromRow)]
pub struct Composer {
    pub id: i64,
    pub name: String,
}

#[derive(Debug, Clone, PartialEq, Eq, PartialOrd, Ord)]
pub struct I64DateTime(pub DateTime<Utc>);

impl TryFrom<i64> for I64DateTime {
    type Error = String;

    fn try_from(value: i64) -> Result<Self, Self::Error> {
        Ok(I64DateTime(
            DateTime::<Utc>::from_timestamp(value, 0).ok_or("Failed converting i64 to DateTime")?,
        ))
    }
}

impl From<&I64DateTime> for i64 {
    fn from(value: &I64DateTime) -> Self {
        value.0.timestamp()
    }
}

#[derive(Debug, Clone, PartialEq, Eq)]
pub struct Pdf {
    pub path: PathBuf,
    pub file_size: u64,
    pub file_hash: String,
}

impl Pdf {
    pub fn get_name(&self) -> &str {
        self.path.file_name().unwrap().to_str().unwrap()
    }

    pub fn validate_own_path(&self) -> std::io::Result<bool> {
        self.validate_path(&self.path)
    }

    pub fn validate_path(&self, path: impl AsRef<Path>) -> std::io::Result<bool> {
        // First compare file size since it is faster than hashing
        let file_size = fs::metadata(path.as_ref())?.len();
        if file_size == self.file_size {
            let file_content = fs::read(path.as_ref())?;
            let file_hash = blake3::hash(&file_content);
            if file_hash.to_string() == self.file_hash {
                return Ok(true);
            }
        }

        Ok(false)
    }
}

impl TryFrom<PathBuf> for Pdf {
    type Error = std::io::Error;

    fn try_from(path: PathBuf) -> Result<Self, Self::Error> {
        let file_size = fs::metadata(path.as_path())?.len();
        let file_content = fs::read(path.as_path())?;
        let file_hash = blake3::hash(&file_content).to_string();

        Ok(Pdf {
            path,
            file_size,
            file_hash,
        })
    }
}
