use gtk::prelude::*;
use std::sync::Arc;

use relm4::{
    component::{AsyncComponent, AsyncComponentParts, Connector},
    gtk::{
        gio::ListStore,
        glib::{self, GString, Type, Value},
        EntryBuffer, EntryCompletion,
    },
    prelude::*,
    AsyncComponentSender,
};
use relm4_components::alert::{Alert, AlertMsg, AlertSettings};

use crate::{database::Database, sheet::Sheet, sheet_dao};

pub struct SheetEditDialogModel {
    database: Arc<Database>,
    hidden: bool,
    sheet: Option<Sheet>,
    name_entry_buffer: EntryBuffer,
    composer_entry_buffer: EntryBuffer,
    composer_entry_completion: EntryCompletion,
    is_book: bool,
    book_sheets: Vec<(String, String, i64)>,
    alert_empty_fields: Connector<Alert>,
}

pub struct SheetEditDialogInit {
    pub database: Arc<Database>,
    pub sheet: Sheet,
}

#[derive(Debug)]
pub enum SheetEditDialogInput {
    Accept,
    Cancel,
}

#[derive(Debug)]
pub enum SheetEditDialogOutput {
    SheetEdited(Sheet),
}

#[relm4::component(pub, async)]
impl AsyncComponent for SheetEditDialogModel {
    type Init = SheetEditDialogInit;
    type Input = SheetEditDialogInput;
    type Output = SheetEditDialogOutput;
    type CommandOutput = ();

    view! {
        gtk::Window {
            #[watch]
            set_visible: !model.hidden,
            set_modal: true,
            set_title: Some("Edit sheet"),
            set_default_width: 10,
            set_default_height: 10,
            gtk::Box {
                set_orientation: gtk::Orientation::Vertical,
                set_margin_all : 15,
                set_spacing: 5,
                gtk::Box {
                    set_spacing: 10,
                    gtk::Label {
                        set_text: "Sheet name"
                    },
                    gtk::Entry {
                        set_buffer: &model.name_entry_buffer,
                        set_width_chars: 40,
                    },
                },
                gtk::Box {
                    set_spacing: 10,
                    gtk::Label {
                        set_text: "Sheet composer"
                    },
                    gtk::Entry {
                        set_buffer: &model.composer_entry_buffer,
                        set_completion: Some(&model.composer_entry_completion),
                        set_hexpand: true,
                    },
                },
                gtk::Box {
                    set_spacing: 10,
                    gtk::Label {
                        set_text: "Book"
                    },
                    gtk::CheckButton {
                        #[watch]
                        set_active: model.is_book,
                    },
                },
                gtk::Box {
                    set_orientation: gtk::Orientation::Horizontal,
                    set_margin_top: 10,
                    set_spacing: 10,
                    set_homogeneous: true,
                    set_halign: gtk::Align::Center,
                    set_hexpand: true,

                    gtk::Button {
                        set_label: "Cancel",
                        connect_clicked[sender] => move |_| sender.input(SheetEditDialogInput::Cancel)
                    },
                    gtk::Button {
                        set_label : "Confirm",
                        connect_clicked[sender] => move |_| sender.input(SheetEditDialogInput::Accept)
                    },
                },
            }
        }
    }

    async fn init(
        params: Self::Init,
        root: Self::Root,
        sender: AsyncComponentSender<Self>,
    ) -> AsyncComponentParts<Self> {
        let sheet = params.sheet;
        let mut sheet_name = String::new();
        let mut sheet_composer = String::new();
        let mut is_book = false;

        match &sheet.kind {
            crate::sheet::SheetKind::Sheet {
                name, composer_id, ..
            } => {
                sheet_name = name.to_string();
                if let Ok(composer) =
                    sheet_dao::get_composer_by_id(&params.database, *composer_id).await
                {
                    sheet_composer = composer.name;
                }
            }
            crate::sheet::SheetKind::Orphan => {
                sheet_name = sheet.pdf.get_name().to_string();
            }
            crate::sheet::SheetKind::Book {
                name,
                composer_id,
                sheet_ids: _,
            } => {
                is_book = true;
                sheet_name = name.to_string();
                if let Ok(composer) =
                    sheet_dao::get_composer_by_id(&params.database, *composer_id).await
                {
                    sheet_composer = composer.name;
                }
                // TODO: load sheets of book
            }
        };

        let composer_entry_completion = EntryCompletion::new();
        let data = [
            "France".to_string(),
            "Italy".to_string(),
            "Sweden".to_string(),
            "Switzerland".to_string(),
        ];
        let store = gtk::ListStore::new(&[glib::Type::STRING]);
        for d in data.iter() {
            store.set(&store.append(), &[(0, &d)]);
        }

        composer_entry_completion.set_model(Some(&store));
        // Use the first (and only) column available to set the autocompletion text
        composer_entry_completion.set_text_column(0);
        // how many keystrokes to wait before attempting to autocomplete?
        composer_entry_completion.set_minimum_key_length(1);
        // whether the completions should be presented in a popup window
        composer_entry_completion.set_popup_completion(true);

        let model = SheetEditDialogModel {
            database: params.database,
            hidden: false,
            sheet: Some(sheet),
            name_entry_buffer: EntryBuffer::new(Some(sheet_name)),
            composer_entry_buffer: EntryBuffer::new(Some(sheet_composer)),
            composer_entry_completion,
            is_book,
            book_sheets: Vec::new(),
            alert_empty_fields: Alert::builder().transient_for(&root).launch(AlertSettings {
                text: String::from("Missing input"),
                secondary_text: Some(String::from("Please make sure all fields are filled")),
                is_modal: true,
                destructive_accept: false,
                confirm_label: Some(String::from("Ok")),
                cancel_label: None,
                option_label: None,
            }),
        };
        let widgets = view_output!();

        AsyncComponentParts { model, widgets }
    }

    // TODO: init_loading_widgets

    async fn update(
        &mut self,
        msg: Self::Input,
        _sender: AsyncComponentSender<Self>,
        _root: &Self::Root,
    ) {
        match msg {
            SheetEditDialogInput::Accept => {
                if let Some(_sheet) = &self.sheet {
                    let sheet_name_string = self.name_entry_buffer.text();
                    let sheet_name = sheet_name_string.trim();

                    let sheet_composer_string = self.composer_entry_buffer.text();
                    let sheet_composer = sheet_composer_string.trim();

                    if sheet_name.is_empty() || sheet_composer.is_empty() {
                        self.alert_empty_fields.emit(AlertMsg::Show);
                        return;
                    }

                    //     match sheet.kind.borrow_mut() {
                    //         crate::sheet::SheetKind::Sheet {
                    //             name,
                    //             composer_id,
                    //             first_page,
                    //             book_id,
                    //         } => {
                    //             todo!("Do something!!!");
                    //             // name = "hello world";
                    //             // name = &mut self.sheet_name.clone();
                    //             // composer_id = 0;
                    //         }
                    //         crate::sheet::SheetKind::Orphan => {
                    //             todo!("Create Sheet");
                    //         }
                    //         crate::sheet::SheetKind::Book {
                    //             name,
                    //             composer_id,
                    //             sheet_ids,
                    //         } => todo!(),
                    //     };
                    // sender
                    //     .output(SheetEditDialogOutput::SheetEdited(sheet))
                    //     .unwrap();
                }
                self.hidden = true;
            }
            SheetEditDialogInput::Cancel => {
                self.hidden = true;
                self.sheet = None;
            }
        }
    }
}
